@@ -18,6 +18,7 @@ package main
 
 import (
 	"context"
+	"errors"
 	"flag"
 	"fmt"
 	"net"
@@ -31,17 +32,19 @@ import (
 	"time"
 
 	"github.com/kelseyhightower/envconfig"
-	"github.com/pkg/errors"
 	"go.opencensus.io/plugin/ochttp"
 	"go.opencensus.io/stats"
 	"go.opencensus.io/stats/view"
 	"go.opencensus.io/trace"
 	"go.uber.org/zap"
+
 	"k8s.io/apimachinery/pkg/types"
 	"k8s.io/apimachinery/pkg/util/wait"
+
 	pkglogging "knative.dev/pkg/logging"
 	"knative.dev/pkg/logging/logkey"
 	"knative.dev/pkg/metrics"
+	pkgnet "knative.dev/pkg/network"
 	"knative.dev/pkg/profiling"
 	"knative.dev/pkg/signals"
 	"knative.dev/pkg/tracing"
@@ -49,7 +52,6 @@ import (
 	"knative.dev/serving/pkg/activator"
 	activatorutil "knative.dev/serving/pkg/activator/util"
 	"knative.dev/serving/pkg/apis/networking"
-	"knative.dev/serving/pkg/autoscaler"
 	pkghttp "knative.dev/serving/pkg/http"
 	"knative.dev/serving/pkg/logging"
 	"knative.dev/serving/pkg/network"
@@ -79,7 +81,6 @@ const (
 	queueDepthN            = "queue_depth"
 
 	healthURLTemplate = "http://127.0.0.1:%d"
-	tcpProbeTimeout   = 100 * time.Millisecond
 	// The 25 millisecond retry interval is an unscientific compromise between wanting to get
 	// started as early as possible while still wanting to give the container some breathing
 	// room to get up and running.
@@ -154,17 +155,18 @@ type config struct {
 }
 
 // Make handler a closure for testing.
-func handler(reqChan chan queue.ReqEvent, breaker *queue.Breaker, handler http.Handler,
-	healthState *health.State, prober func() bool, isAggressive bool) func(http.ResponseWriter, *http.Request) {
+func proxyHandler(reqChan chan queue.ReqEvent, breaker *queue.Breaker, tracingEnabled bool, next http.Handler) http.HandlerFunc {
 	return func(w http.ResponseWriter, r *http.Request) {
-		// TODO: Move probe part to network.NewProbeHandler if possible or another handler.
-		if ph := network.KnativeProbeHeader(r); ph != "" {
-			handleKnativeProbe(w, r, ph, healthState, prober, isAggressive)
+		if network.IsKubeletProbe(r) {
+			next.ServeHTTP(w, r)
 			return
 		}
 
-		proxyCtx, proxySpan := trace.StartSpan(r.Context(), "proxy")
-		defer proxySpan.End()
+		if tracingEnabled {
+			proxyCtx, proxySpan := trace.StartSpan(r.Context(), "proxy")
+			r = r.WithContext(proxyCtx)
+			defer proxySpan.End()
+		}
 
 		// Metrics for autoscaling.
 		in, out := queue.ReqIn, queue.ReqOut
@@ -180,7 +182,7 @@ func handler(reqChan chan queue.ReqEvent, breaker *queue.Breaker, handler http.H
 		// Enforce queuing and concurrency limits.
 		if breaker != nil {
 			if err := breaker.Maybe(r.Context(), func() {
-				handler.ServeHTTP(w, r.WithContext(proxyCtx))
+				next.ServeHTTP(w, r)
 			}); err != nil {
 				switch err {
 				case context.DeadlineExceeded, queue.ErrRequestQueueFull:
@@ -190,37 +192,49 @@ func handler(reqChan chan queue.ReqEvent, breaker *queue.Breaker, handler http.H
 				}
 			}
 		} else {
-			handler.ServeHTTP(w, r.WithContext(proxyCtx))
+			next.ServeHTTP(w, r)
 		}
 	}
 }
 
-func handleKnativeProbe(w http.ResponseWriter, r *http.Request, ph string, healthState *health.State, prober func() bool, isAggressive bool) {
-	_, probeSpan := trace.StartSpan(r.Context(), "probe")
-	defer probeSpan.End()
+func knativeProbeHandler(healthState *health.State, prober func() bool, isAggressive bool, tracingEnabled bool, next http.Handler) http.HandlerFunc {
+	return func(w http.ResponseWriter, r *http.Request) {
+		ph := network.KnativeProbeHeader(r)
 
-	if ph != queue.Name {
-		http.Error(w, fmt.Sprintf(badProbeTemplate, ph), http.StatusBadRequest)
-		probeSpan.Annotate([]trace.Attribute{
-			trace.StringAttribute("queueproxy.probe.error", fmt.Sprintf(badProbeTemplate, ph))}, "error")
-		return
-	}
-
-	if prober == nil {
-		http.Error(w, "no probe", http.StatusInternalServerError)
-		probeSpan.Annotate([]trace.Attribute{
-			trace.StringAttribute("queueproxy.probe.error", "no probe")}, "error")
-		return
-	}
-
-	healthState.HandleHealthProbe(func() bool {
-		if !prober() {
-			probeSpan.Annotate([]trace.Attribute{
-				trace.StringAttribute("queueproxy.probe.error", "container not ready")}, "error")
-			return false
+		if ph == "" {
+			next.ServeHTTP(w, r)
+			return
 		}
-		return true
-	}, isAggressive, w)
+
+		var probeSpan *trace.Span
+		if tracingEnabled {
+			_, probeSpan = trace.StartSpan(r.Context(), "probe")
+			defer probeSpan.End()
+		}
+
+		if ph != queue.Name {
+			http.Error(w, fmt.Sprintf(badProbeTemplate, ph), http.StatusBadRequest)
+			probeSpan.Annotate([]trace.Attribute{
+				trace.StringAttribute("queueproxy.probe.error", fmt.Sprintf(badProbeTemplate, ph))}, "error")
+			return
+		}
+
+		if prober == nil {
+			http.Error(w, "no probe", http.StatusInternalServerError)
+			probeSpan.Annotate([]trace.Attribute{
+				trace.StringAttribute("queueproxy.probe.error", "no probe")}, "error")
+			return
+		}
+
+		healthState.HandleHealthProbe(func() bool {
+			if !prober() {
+				probeSpan.Annotate([]trace.Attribute{
+					trace.StringAttribute("queueproxy.probe.error", "container not ready")}, "error")
+				return false
+			}
+			return true
+		}, isAggressive, w)
+	}
 }
 
 func probeQueueHealthPath(port int, timeoutSeconds int) error {
@@ -322,30 +336,20 @@ func main() {
 	}
 
 	// Setup reporters and processes to handle stat reporting.
-	promStatReporter, err := queue.NewPrometheusStatsReporter(env.ServingNamespace, env.ServingConfiguration, env.ServingRevision, env.ServingPod, reportingPeriod)
+	promStatReporter, err := queue.NewPrometheusStatsReporter(
+		env.ServingNamespace, env.ServingConfiguration, env.ServingRevision,
+		env.ServingPod, reportingPeriod)
 	if err != nil {
 		logger.Fatalw("Failed to create stats reporter", zap.Error(err))
 	}
 
-	statChan := make(chan autoscaler.Stat)
-	defer close(statChan)
-	go func() {
-		for s := range statChan {
-			promStatReporter.Report(s)
-		}
-	}()
-
 	reqChan := make(chan queue.ReqEvent, requestCountingQueueLength)
 	defer close(reqChan)
 
 	reportTicker := time.NewTicker(reportingPeriod)
 	defer reportTicker.Stop()
 
-	queue.NewStats(env.ServingPod, queue.Channels{
-		ReqChan:    reqChan,
-		ReportChan: reportTicker.C,
-		StatChan:   statChan,
-	}, time.Now())
+	queue.NewStats(time.Now(), reqChan, reportTicker.C, promStatReporter.Report)
 
 	// Setup probe to run for checking user-application healthiness.
 	probe := buildProbe(env.ServingReadinessProbe)
@@ -448,13 +452,14 @@ func buildServer(env config, healthState *health.State, rp *readiness.Probe, req
 
 	httpProxy := httputil.NewSingleHostReverseProxy(target)
 	httpProxy.Transport = buildTransport(env, logger)
-	httpProxy.ErrorHandler = network.ErrorHandler(logger)
-
+	httpProxy.ErrorHandler = pkgnet.ErrorHandler(logger)
+	httpProxy.BufferPool = network.NewBufferPool()
 	httpProxy.FlushInterval = -1
 	activatorutil.SetupHeaderPruning(httpProxy)
 
 	breaker := buildBreaker(env)
 	metricsSupported := supportsMetrics(env, logger)
+	tracingEnabled := env.TracingConfigBackend != tracingconfig.None
 
 	// Create queue handler chain.
 	// Note: innermost handlers are specified first, ie. the last handler in the chain will be executed first.
@@ -463,7 +468,7 @@ func buildServer(env config, healthState *health.State, rp *readiness.Probe, req
 		composedHandler = pushRequestMetricHandler(httpProxy, appRequestCountM, appResponseTimeInMsecM,
 			queueDepthM, breaker, env)
 	}
-	composedHandler = http.HandlerFunc(handler(reqChan, breaker, composedHandler, healthState, rp.ProbeContainer, rp.IsAggressive()))
+	composedHandler = proxyHandler(reqChan, breaker, tracingEnabled, composedHandler)
 	composedHandler = queue.ForwardedShimHandler(composedHandler)
 	composedHandler = queue.TimeToFirstByteTimeoutHandler(composedHandler,
 		time.Duration(env.RevisionTimeoutSeconds)*time.Second, "request timeout")
@@ -474,14 +479,16 @@ func buildServer(env config, healthState *health.State, rp *readiness.Probe, req
 			nil /*queueDepthM*/, nil /*breaker*/, env)
 	}
 	composedHandler = tracing.HTTPSpanMiddleware(composedHandler)
+
+	composedHandler = knativeProbeHandler(healthState, rp.ProbeContainer, rp.IsAggressive(), tracingEnabled, composedHandler)
 	composedHandler = network.NewProbeHandler(composedHandler)
 
-	return network.NewServer(":"+strconv.Itoa(env.QueueServingPort), composedHandler)
+	return pkgnet.NewServer(":"+strconv.Itoa(env.QueueServingPort), composedHandler)
 }
 
 func buildTransport(env config, logger *zap.SugaredLogger) http.RoundTripper {
 	if env.TracingConfigBackend == tracingconfig.None {
-		return network.AutoTransport
+		return pkgnet.AutoTransport
 	}
 
 	oct := tracing.NewOpenCensusTracer(tracing.WithExporter(env.ServingPod, logger))
@@ -494,7 +501,7 @@ func buildTransport(env config, logger *zap.SugaredLogger) http.RoundTripper {
 	})
 
 	return &ochttp.Transport{
-		Base: network.AutoTransport,
+		Base: pkgnet.AutoTransport,
 	}
 }
 
